@@ -13,7 +13,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * SCCS: @(#) tclIOUtil.c 1.123 96/04/29 14:08:24
+ * SCCS: @(#) tclIOUtil.c 1.128 96/10/02 12:25:36
  */
 
 #include "tclInt.h"
@@ -41,8 +41,9 @@ static Detached *detList = NULL;	/* List of all detached proceses. */
  */
 
 static Tcl_File	FileForRedirect _ANSI_ARGS_((Tcl_Interp *interp,
-	                    char *spec, int atOk, char *arg, int flags,
-	                    char *nextArg, int *skipPtr, int *closePtr));
+	                    char *spec, int atOk, char *arg, char *nextArg, 
+			    int flags, int *skipPtr, int *closePtr,
+			    Tcl_DString *namePtr));
 
 /*
  *----------------------------------------------------------------------
@@ -66,53 +67,54 @@ static Tcl_File	FileForRedirect _ANSI_ARGS_((Tcl_Interp *interp,
  */
 
 static Tcl_File
-FileForRedirect(interp, spec, atOk, arg, flags, nextArg, skipPtr, closePtr)
-    Tcl_Interp *interp;			/* Intepreter to use for error
-					 * reporting. */
-    register char *spec;			/* Points to character just after
-					 * redirection character. */
-    int atOk;				/* Non-zero means '@' notation is
-					 * OK, zero means it isn't. */
-    char *arg;				/* Pointer to entire argument
-					 * containing spec:  used for error
-					 * reporting. */
-    int flags;				/* Flags to use for opening file. */
-    char *nextArg;			/* Next argument in argc/argv
-					 * array, if needed for file name.
-					 * May be NULL. */
-    int *skipPtr;			/* This value is incremented if
-					 * nextArg is used for redirection
-					 * spec. */
-    int *closePtr;			/* This value is set to 1 if the file
-					 * that's returned must be closed, 0
-					 * if it was specified with "@" so
-					 * it must be left open. */
+FileForRedirect(interp, spec, atOK, arg, nextArg, flags, skipPtr, closePtr, 
+	namePtr)
+    Tcl_Interp *interp;		/* Intepreter to use for error reporting. */
+    char *spec;			/* Points to character just after
+				 * redirection character. */
+    char *arg;			/* Pointer to entire argument containing 
+				 * spec:  used for error reporting. */
+    int atOK;			/* Non-zero means that '@' notation can be 
+				 * used to specify a channel, zero means that
+				 * it isn't. */
+    char *nextArg;		/* Next argument in argc/argv array, if needed 
+				 * for file name or channel name.  May be 
+				 * NULL. */
+    int flags;			/* Flags to use for opening file or to 
+				 * specify mode for channel. */
+    int *skipPtr;		/* Filled with 1 if redirection target was
+				 * in spec, 2 if it was in nextArg. */
+    int *closePtr;		/* Filled with one if the caller should 
+				 * close the file when done with it, zero
+				 * otherwise. */
+    Tcl_DString *namePtr;	/* Pointer to an initialized Tcl_DString that
+				 * is filled with the name of the file that
+				 * was opened.  Unmodified if spec refers 
+				 * to a channel.  */
 {
     int writing = (flags & O_WRONLY);
     Tcl_Channel chan;
     Tcl_File file;
 
-    if (atOk && (*spec == '@')) {
+    *skipPtr = 1;
+    if ((atOK != 0)  && (*spec == '@')) {
 	spec++;
-	if (*spec == 0) {
+	if (*spec == '\0') {
 	    spec = nextArg;
 	    if (spec == NULL) {
 		goto badLastArg;
 	    }
-	    *skipPtr += 1;
+	    *skipPtr = 2;
 	}
         chan = Tcl_GetChannel(interp, spec, NULL);
         if (chan == (Tcl_Channel) NULL) {
             return NULL;
         }
-	*closePtr = 0;
         file = Tcl_GetChannelFile(chan, writing ? TCL_WRITABLE : TCL_READABLE);
         if (file == NULL) {
-            Tcl_AppendResult(interp,
-                    "channel \"",
-                    Tcl_GetChannelName(chan),
+            Tcl_AppendResult(interp, "channel \"", Tcl_GetChannelName(chan),
                     "\" wasn't opened for ",
-                    writing ? "writing" : "reading", (char *) NULL);
+                    ((writing) ? "writing" : "reading"), (char *) NULL);
             return NULL;
         }
 	if (writing) {
@@ -126,30 +128,29 @@ FileForRedirect(interp, spec, atOk, arg, flags, nextArg, skipPtr, closePtr)
             Tcl_Flush(chan);
 	}
     } else {
-	Tcl_DString buffer;
 	char *name;
 
-	if (*spec == 0) {
+	if (*spec == '\0') {
 	    spec = nextArg;
 	    if (spec == NULL) {
 		goto badLastArg;
 	    }
-	    *skipPtr += 1;
+	    *skipPtr = 2;
 	}
-	name = Tcl_TranslateFileName(interp, spec, &buffer);
-	if (name) {
+	name = Tcl_TranslateFileName(interp, spec, namePtr);
+	if (name != NULL) {
 	    file = TclOpenFile(name, flags);
 	} else {
 	    file = NULL;
 	}
-	Tcl_DStringFree(&buffer);
 	if (file == NULL) {
 	    Tcl_AppendResult(interp, "couldn't ",
-		    (writing) ? "write" : "read", " file \"", spec, "\": ",
+		    ((writing) ? "write" : "read"), " file \"", spec, "\": ",
 		    Tcl_PosixError(interp), (char *) NULL);
+	    Tcl_DStringFree(namePtr);
 	    return NULL;
 	}
-	*closePtr = 1;
+        *closePtr = 1;
     }
     return file;
 
@@ -665,7 +666,7 @@ TclCleanupChildren(interp, numPids, pidPtr, errorChan)
             }
         }
         
-	Tcl_Close(NULL, errorChan);
+	Tcl_Close((Tcl_Interp *) NULL, errorChan);
     }
 
     /*
@@ -757,41 +758,39 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 				 * at *pidPtr right now. */
     int cmdCount;		/* Count of number of distinct commands
 				 * found in argc/argv. */
-    char *input = NULL;		/* If non-null, then this points to a
+    char *inputLiteral = NULL;	/* If non-null, then this points to a
 				 * string containing input data (specified
 				 * via <<) to be piped to the first process
 				 * in the pipeline. */
-    Tcl_File inputFile = NULL;
-				/* If != NULL, gives file to use as input for
+    Tcl_File inputFile = NULL;	/* If != NULL, gives file to use as input for
 				 * first process in pipeline (specified via <
 				 * or <@). */
-    int closeInput = 0;		/* If non-zero, then must close inputId
-				 * when cleaning up (zero means the file needs
-				 * to stay open for some other reason). */
-    Tcl_File outputFile = NULL;
-				/* Writable file for output from last command
+    Tcl_DString inputFileName;	/* If non-empty, gives name of file that 
+				 * corresponds to inputFile. */
+    int inputClose = 0;		/* If non-zero, then inputFile should be 
+    				 * closed when cleaning up. */
+    Tcl_File outputFile = NULL;	/* Writable file for output from last command
 				 * in pipeline (could be file or pipe).  NULL
 				 * means use stdout. */
-    int closeOutput = 0;	/* Non-zero means must close outputId when
-				 * cleaning up (similar to closeInput). */
-    Tcl_File errorFile = NULL;
-				/* Writable file for error output from all
+    Tcl_DString outputFileName;	/* If non-empty, gives name of file that 
+				 * corresponds to outputFile. */
+    int outputClose = 0;	/* If non-zero, then outputFile should be 
+    				 * closed when cleaning up. */
+    Tcl_File errorFile = NULL;	/* Writable file for error output from all
 				 * commands in pipeline.  NULL means use
 				 * stderr. */
-    int closeError = 0;		/* Non-zero means must close errorId when
-				 * cleaning up. */
-    int skip;			/* Number of arguments to skip (because they
-				 * specify redirection). */
-    int lastBar;
-    int i, j;
+    Tcl_DString errorFileName;	/* If non-empty, gives name of file that 
+				 * corresponds to errorFile. */
+    int errorClose = 0;		/* If non-zero, then errorFile should be 
+    				 * closed when cleaning up. */
     char *p;
-    int hasPipes = TclHasPipes();
-    char finalOut[L_tmpnam];
-    char intIn[L_tmpnam];
+    int skip, lastBar, lastArg, i, j, atOK, flags, errorToOutput;
+    Tcl_DString execBuffer;
+    Tcl_File pipeIn;
+    Tcl_File curInFile, curOutFile, curErrFile;
+    char *curInFileName, *curOutFileName, *curErrFileName;
+    Tcl_Channel channel;
 
-    finalOut[0]  = '\0';
-    intIn[0] = '\0';
-    
     if (inPipePtr != NULL) {
 	*inPipePtr = NULL;
     }
@@ -802,37 +801,65 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 	*errFilePtr = NULL;
     }
 
+    Tcl_DStringInit(&inputFileName);
+    Tcl_DStringInit(&outputFileName);
+    Tcl_DStringInit(&errorFileName);
+    Tcl_DStringInit(&execBuffer);
+    
+    pipeIn = NULL;
+    curInFile = NULL;
+    curOutFile = NULL;
+    curErrFile = NULL;
+
+    numPids = 0;
+    pidPtr = NULL;
+
     /*
      * First, scan through all the arguments to figure out the structure
      * of the pipeline.  Process all of the input and output redirection
      * arguments and remove them from the argument list in the pipeline.
      * Count the number of distinct processes (it's the number of "|"
-     * arguments plus one) but don't remove the "|" arguments.
+     * arguments plus one) but don't remove the "|" arguments because 
+     * they'll be used in the second pass to seperate the individual 
+     * child processes.  Cannot start the child processes in this pass 
+     * because the redirection symbols may appear anywhere in the 
+     * command line -- e.g., the '<' that specifies the input to the 
+     * entire pipe may appear at the very end of the argument list.
      */
 
-    cmdCount = 1;
     lastBar = -1;
+    cmdCount = 1;
     for (i = 0; i < argc; i++) {
-	if ((argv[i][0] == '|') && (((argv[i][1] == 0))
-		|| ((argv[i][1] == '&') && (argv[i][2] == 0)))) {
-	    if ((i == (lastBar+1)) || (i == (argc-1))) {
-		interp->result = "illegal use of | or |& in command";
-		return -1;
+        skip = 0;
+	p = argv[i];
+	switch (*p++) {
+	case '|':
+	    if (*p == '&') {
+		p++;
+	    }
+	    if (*p == '\0') {
+		if ((i == (lastBar + 1)) || (i == (argc - 1))) {
+		    interp->result = "illegal use of | or |& in command";
+		    goto error;
+		}
 	    }
 	    lastBar = i;
 	    cmdCount++;
-	    continue;
-	} else if (argv[i][0] == '<') {
-	    if ((inputFile != NULL) && closeInput) {
+	    break;
+
+	case '<':
+	    if (inputClose != 0) {
+		inputClose = 0;
+		Tcl_DStringFree(&inputFileName);
 		TclCloseFile(inputFile);
 	    }
-	    inputFile = NULL;
-	    skip = 1;
-	    if (argv[i][1] == '<') {
-		input = argv[i]+2;
-		if (*input == 0) {
-		    input = argv[i+1];
-		    if (input == 0) {
+	    if (*p == '<') {
+		inputFile = NULL;
+		inputLiteral = p + 1;
+		skip = 1;
+		if (*inputLiteral == '\0') {
+		    inputLiteral = argv[i + 1];
+		    if (inputLiteral == NULL) {
 			Tcl_AppendResult(interp, "can't specify \"", argv[i],
 				"\" as last word in command", (char *) NULL);
 			goto error;
@@ -840,215 +867,294 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
 		    skip = 2;
 		}
 	    } else {
-		input = 0;
-		inputFile = FileForRedirect(interp, argv[i]+1, 1, argv[i],
-			O_RDONLY, argv[i+1], &skip, &closeInput);
+		inputLiteral = NULL;
+		inputFile = FileForRedirect(interp, p, 1, argv[i], 
+			argv[i + 1], O_RDONLY, &skip, &inputClose, 
+			&inputFileName);
 		if (inputFile == NULL) {
 		    goto error;
 		}
+	    }
+	    break;
 
-		/* When Win32s dies out, this code can be removed */
-		if (!hasPipes) {
-		    if (!closeInput) {
-			Tcl_AppendResult(interp, "redirection with '@'",
-				" notation is not supported on this system",
-				(char *) NULL);
-			goto error;
-		    }
-		    strcpy(intIn, skip == 1 ? argv[i]+1 : argv[i+1]);
-		}
-	    }
-	} else if (argv[i][0] == '>') {
-	    int append, useForStdErr, useForStdOut, mustClose, atOk, flags;
-	    Tcl_File file;
-
-	    skip = atOk = 1;
-	    append = useForStdErr = 0;
-	    useForStdOut = 1;
-	    if (argv[i][1] == '>') {
-		p = argv[i] + 2;
-		append = 1;
-		atOk = 0;
-		flags = O_WRONLY|O_CREAT;
-	    } else {
-		p = argv[i] + 1;
-		flags = O_WRONLY|O_CREAT|O_TRUNC;
-	    }
-	    if (*p == '&') {
-		useForStdErr = 1;
-		p++;
-	    }
-	    file = FileForRedirect(interp, p, atOk, argv[i], flags, argv[i+1],
-		    &skip, &mustClose);
-	    if (file == NULL) {
-		goto error;
-	    }
-
-	    /* When Win32s dies out, this code can be removed */
-	    if (!hasPipes) {
-		if (!mustClose) {
-		    Tcl_AppendResult(interp, "redirection with '@'",
-			    " notation is not supported on this system",
-			    (char *) NULL);
-		    goto error;
-		}
-		strcpy(finalOut, skip == 1 ? p : argv[i+1]);
-	    }
-
-	    if (hasPipes && append) {
-		TclSeekFile(file, 0L, 2);
-	    }
-
-	    /*
-	     * Got the file descriptor.  Now use it for standard output,
-	     * standard error, or both, depending on the redirection.
-	     */
-
-	    if (useForStdOut) {
-		if ((outputFile != NULL) && closeOutput) {
-		    TclCloseFile(outputFile);
-		}
-		outputFile = file;
-		closeOutput = mustClose;
-	    }
-	    if (useForStdErr) {
-		if ((errorFile != NULL) && closeError) {
-		    TclCloseFile(errorFile);
-		}
-		errorFile = file;
-		closeError = (useForStdOut) ? 0 : mustClose;
-	    }
-	} else if ((argv[i][0] == '2') && (argv[i][1] == '>')) {
-	    int append, atOk, flags;
-
-	    if ((errorFile != NULL) && closeError) {
-		TclCloseFile(errorFile);
-	    }
-	    skip = 1;
-	    p = argv[i] + 2;
+	case '>':
+	    atOK = 1;
+	    flags = O_WRONLY | O_CREAT | O_TRUNC;
+	    errorToOutput = 0;
 	    if (*p == '>') {
 		p++;
-		append = 1;
-		atOk = 0;
-		flags = O_WRONLY|O_CREAT;
-	    } else {
-		append = 0;
-		atOk = 1;
-		flags = O_WRONLY|O_CREAT|O_TRUNC;
+		atOK = 0;
+		flags = O_WRONLY | O_CREAT;
 	    }
-	    errorFile = FileForRedirect(interp, p, atOk, argv[i], flags,
-		    argv[i+1], &skip, &closeError);
+	    if (*p == '&') {
+		if (errorClose != 0) {
+		    errorClose = 0;
+		    Tcl_DStringFree(&errorFileName);
+		    TclCloseFile(errorFile);
+		}
+		errorToOutput = 1;
+		p++;
+	    }
+
+	    if (outputClose != 0) {
+		outputClose = 0;
+		Tcl_DStringFree(&outputFileName);
+		TclCloseFile(outputFile);
+	    }
+	    outputFile = FileForRedirect(interp, p, atOK, argv[i], 
+		    argv[i + 1], flags, &skip, &outputClose, 
+		    &outputFileName);
+	    if (outputFile == NULL) {
+		goto error;
+	    }
+	    if (atOK == 0) {
+		TclSeekFile(outputFile, 0, SEEK_END);
+	    }
+	    if (errorToOutput) {
+		errorClose = 0;
+		errorFile = outputFile;
+	    }
+	    break;
+
+	case '2':
+	    if (*p != '>') {
+		break;
+	    }
+	    p++;
+	    atOK = 1;
+	    flags = O_WRONLY | O_CREAT | O_TRUNC;
+	    if (*p == '>') {
+		p++;
+		atOK = 0;
+		flags = O_WRONLY | O_CREAT;
+	    }
+	    if (errorClose != 0) {
+		errorClose = 0;
+		Tcl_DStringFree(&errorFileName);
+		TclCloseFile(errorFile);
+	    }
+	    errorFile = FileForRedirect(interp, p, atOK, argv[i], 
+		    argv[i + 1], flags, &skip, &errorClose, 
+		    &errorFileName);
 	    if (errorFile == NULL) {
 		goto error;
 	    }
-	    if (hasPipes && append) {
-		TclSeekFile(errorFile, 0L, 2);
+	    if (atOK == 0) {
+		TclSeekFile(errorFile, 0, SEEK_END);
 	    }
-	} else {
-	    continue;
+	    break;
 	}
-	for (j = i+skip; j < argc; j++) {
-	    argv[j-skip] = argv[j];
+
+	if (skip != 0) {
+	    for (j = i + skip; j < argc; j++) {
+		argv[j - skip] = argv[j];
+	    }
+	    argc -= skip;
+	    i -= 1;
 	}
-	argc -= skip;
-	i -= 1;			/* Process next arg from same position. */
-    }
-    if (argc == 0) {
-	interp->result =  "didn't specify command to execute";
-	return -1;
     }
 
-    if ((hasPipes && inputFile == NULL) || (!hasPipes && intIn[0] == '\0')) {
-	if (input != NULL) {
-
+    if (inputFile == NULL) {
+	if (inputLiteral != NULL) {
 	    /*
 	     * The input for the first process is immediate data coming from
 	     * Tcl.  Create a temporary file for it and put the data into the
 	     * file.
 	     */
-	    
-	    inputFile = TclCreateTempFile(input);
-	    closeInput = 1;
+	    inputFile = TclCreateTempFile(inputLiteral, &inputFileName);
 	    if (inputFile == NULL) {
 		Tcl_AppendResult(interp,
 			"couldn't create input file for command: ",
 			Tcl_PosixError(interp), (char *) NULL);
 		goto error;
 	    }
+	    inputClose = 1;
 	} else if (inPipePtr != NULL) {
-	    Tcl_File inPipe, outPipe;
 	    /*
 	     * The input for the first process in the pipeline is to
-	     * come from a pipe that can be written from this end.
+	     * come from a pipe that can be written from by the caller.
 	     */
 
-	    if (!hasPipes || TclCreatePipe(&inPipe, &outPipe) == 0) {
-		Tcl_AppendResult(interp,
+	    if (TclCreatePipe(&inputFile, inPipePtr) == 0) {
+		Tcl_AppendResult(interp, 
 			"couldn't create input pipe for command: ",
 			Tcl_PosixError(interp), (char *) NULL);
 		goto error;
 	    }
-	    inputFile = inPipe;
-	    closeInput = 1;
-	    *inPipePtr = outPipe;
+	    inputClose = 1;
+	} else {
+	    /*
+	     * The input for the first process comes from stdin.
+	     */
+
+	    channel = Tcl_GetStdChannel(TCL_STDIN);
+	    if (channel != NULL) {
+		inputFile = Tcl_GetChannelFile(channel, TCL_READABLE);
+	    }
 	}
     }
 
-    /*
-     * Set up a pipe to receive output from the pipeline, if no other
-     * output sink has been specified.
-     */
+    if (outputFile == NULL) {
+	if (outPipePtr != NULL) {
+	    /*
+	     * Output from the last process in the pipeline is to go to a
+	     * pipe that can be read by the caller.
+	     */
 
-    if ((outputFile == NULL) && (outPipePtr != NULL)) {
-	if (!hasPipes) {
-	    tmpnam(finalOut);
-	} else {
-	    Tcl_File inPipe, outPipe;
-	    if (TclCreatePipe(&inPipe, &outPipe) == 0) {
-		Tcl_AppendResult(interp,
+	    if (TclCreatePipe(outPipePtr, &outputFile) == 0) {
+		Tcl_AppendResult(interp, 
 			"couldn't create output pipe for command: ",
 			Tcl_PosixError(interp), (char *) NULL);
 		goto error;
 	    }
-	    outputFile = outPipe;
-	    closeOutput = 1;
-	    *outPipePtr = inPipe;
+	    outputClose = 1;
+	} else {
+	    /*
+	     * The output for the last process goes to stdout.
+	     */
+
+	    channel = Tcl_GetStdChannel(TCL_STDOUT);
+	    if (channel) {
+		outputFile = Tcl_GetChannelFile(channel, TCL_WRITABLE);
+	    }
 	}
     }
 
-    /*
-     * Set up the standard error output sink for the pipeline, if
-     * requested.  Use a temporary file which is opened, then deleted.
-     * Could potentially just use pipe, but if it filled up it could
-     * cause the pipeline to deadlock:  we'd be waiting for processes
-     * to complete before reading stderr, and processes couldn't complete
-     * because stderr was backed up.
-     */
+    if (errorFile == NULL) {
+	if (errFilePtr != NULL) {
+	    /*
+	     * Set up the standard error output sink for the pipeline, if
+	     * requested.  Use a temporary file which is opened, then deleted.
+	     * Could potentially just use pipe, but if it filled up it could
+	     * cause the pipeline to deadlock:  we'd be waiting for processes
+	     * to complete before reading stderr, and processes couldn't 
+	     * complete because stderr was backed up.
+	     */
 
-    if (errFilePtr && !errorFile) {
-	*errFilePtr = TclCreateTempFile(NULL);
-	if (*errFilePtr == NULL) {
-	    Tcl_AppendResult(interp,
-		    "couldn't create error file for command: ",
-		    Tcl_PosixError(interp), (char *) NULL);
-	    goto error;
+	    errorFile = TclCreateTempFile(NULL, &errorFileName);
+	    if (errorFile == NULL) {
+		Tcl_AppendResult(interp,
+			"couldn't create error file for command: ",
+			Tcl_PosixError(interp), (char *) NULL);
+		goto error;
+	    }
+	    *errFilePtr = errorFile;
+	} else {
+	    /*
+	     * Errors from the pipeline go to stderr.
+	     */
+
+	    channel = Tcl_GetStdChannel(TCL_STDERR);
+	    if (channel) {
+		errorFile = Tcl_GetChannelFile(channel, TCL_WRITABLE);
+	    }
 	}
-	errorFile = *errFilePtr;
-	closeError = 0;
     }
 	
     /*
-     * Scan through the argc array, forking off a process for each
-     * group of arguments between "|" arguments.
+     * Scan through the argc array, creating a process for each
+     * group of arguments between the "|" characters.
      */
 
-    pidPtr = (int *) ckalloc((unsigned) (cmdCount * sizeof(int)));
     Tcl_ReapDetachedProcs();
+    pidPtr = (int *) ckalloc((unsigned) (cmdCount * sizeof(int)));
 
-    if (TclSpawnPipeline(interp, pidPtr, &numPids, argc, argv, 
-	    inputFile, outputFile, errorFile, intIn, finalOut) == 0) {
-	goto error;
+    curInFile = inputFile;
+    curInFileName = Tcl_DStringValue(&inputFileName);
+    if (curInFileName[0] == '\0') {
+	curInFileName = NULL;
     }
+
+    for (i = 0; i < argc; i = lastArg + 1) { 
+	int joinThisError, pid;
+
+	/*
+	 * Convert the program name into native form. 
+	 */
+
+	argv[i] = Tcl_TranslateFileName(interp, argv[i], &execBuffer);
+	if (argv[i] == NULL) {
+	    goto error;
+	}
+
+	/*
+	 * Find the end of the current segment of the pipeline.
+	 */
+
+	joinThisError = 0;
+	for (lastArg = i; lastArg < argc; lastArg++) {
+	    if (argv[lastArg][0] == '|') { 
+		if (argv[lastArg][1] == '\0') { 
+		    break;
+		}
+		if ((argv[lastArg][1] == '&') && (argv[lastArg][2] == '\0')) {
+		    joinThisError = 1;
+		    break;
+		}
+	    }
+	}
+	argv[lastArg] = NULL;
+
+	/*
+	 * If this is the last segment, use the specified outputFile.
+	 * Otherwise create an intermediate pipe.  pipeIn will become the
+	 * curInFile for the next segment of the pipe.
+	 */
+
+	if (lastArg == argc) { 
+	    curOutFile = outputFile;
+	    curOutFileName = Tcl_DStringValue(&outputFileName);
+	    if (curOutFileName[0] == '\0') {
+		curOutFileName = NULL;
+	    }
+	} else {
+	    if (TclCreatePipe(&pipeIn, &curOutFile) == 0) {
+		Tcl_AppendResult(interp, "couldn't create pipe: ",
+			Tcl_PosixError(interp), (char *) NULL);
+		goto error;
+	    }
+	    curOutFileName = NULL;
+	}
+
+	if (joinThisError != 0) {
+	    curErrFile = curOutFile;
+	    curErrFileName = curOutFileName;
+	} else {
+	    curErrFile = errorFile;
+	    curErrFileName = Tcl_DStringValue(&errorFileName);
+	    if (curErrFileName[0] == '\0') {
+		curErrFileName = NULL;
+	    }
+	}
+
+	if (TclpCreateProcess(interp, lastArg - i, argv + i,
+		curInFile, curOutFile, curErrFile, curInFileName, 
+		curOutFileName, curErrFileName, &pid) != TCL_OK) {
+	    goto error;
+	}
+	Tcl_DStringFree(&execBuffer);
+
+	pidPtr[numPids] = pid;
+	numPids++;
+
+	/*
+	 * Close off our copies of file descriptors that were set up for
+	 * this child, then set up the input for the next child.
+	 */
+
+	if ((curInFile != NULL) && (curInFile != inputFile)) {
+	    TclCloseFile(curInFile);
+	}
+	curInFile = pipeIn;
+	curInFileName = NULL;
+	pipeIn = NULL;
+
+	if ((curOutFile != NULL) && (curOutFile != outputFile)) {
+	    TclCloseFile(curOutFile);
+	}
+	curOutFile = NULL;
+    }
+
     *pidArrayPtr = pidPtr;
 
     /*
@@ -1056,13 +1162,18 @@ TclCreatePipeline(interp, argc, argv, pidArrayPtr, inPipePtr,
      */
 
 cleanup:
-    if ((inputFile != NULL) && closeInput) {
+    Tcl_DStringFree(&inputFileName);
+    Tcl_DStringFree(&outputFileName);
+    Tcl_DStringFree(&errorFileName);
+    Tcl_DStringFree(&execBuffer);
+
+    if (inputClose) {
 	TclCloseFile(inputFile);
     }
-    if ((outputFile != NULL) && closeOutput) {
+    if (outputClose) {
 	TclCloseFile(outputFile);
     }
-    if ((errorFile != NULL) && closeError) {
+    if (errorClose) {
 	TclCloseFile(errorFile);
     }
     return numPids;
@@ -1074,6 +1185,15 @@ cleanup:
      */
 
 error:
+    if (pipeIn != NULL) {
+	TclCloseFile(pipeIn);
+    }
+    if ((curOutFile != NULL) && (curOutFile != outputFile)) {
+	TclCloseFile(curOutFile);
+    }
+    if ((curInFile != NULL) && (curInFile != inputFile)) {
+	TclCloseFile(curInFile);
+    }
     if ((inPipePtr != NULL) && (*inPipePtr != NULL)) {
 	TclCloseFile(*inPipePtr);
 	*inPipePtr = NULL;
